import numpy as np
import theano.tensor as tt


def probit(x):
    """Probit function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = 0.5 + 0.5 * tt.erf(x / tt.sqrt(2))
    result = tt.switch(tt.eq(result, 0), eps, result)
    result = tt.switch(tt.eq(result, 1), 1 - eps, result)

    return result


def cloglog(x):
    """Cloglog function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = 1 - tt.exp(-tt.exp(x))
    result = tt.switch(tt.eq(result, 0), eps, result)
    result = tt.switch(tt.eq(result, 1), 1 - eps, result)

    return result


def logit(x):
    """Logit function that ensures result is in (0, 1)"""
    eps = np.finfo(float).eps
    result = tt.nnet.sigmoid(x)
    result = tt.switch(tt.eq(result, 0), eps, result)
    result = tt.switch(tt.eq(result, 1), 1 - eps, result)
    return result


def identity(x):
    return x


def inverse_squared(x):
    return tt.inv(tt.sqrt(x))



# pylint: disable=unused-argument
import numpy as np

from .family import Family


class MultivariateFamily(Family):
    def predict(self, model, posterior, linear_predictor):
        """Predict mean response"""
        mean = self.link.linkinv(linear_predictor)
        obs_n = mean.shape[-1]
        name = model.response.name + "_mean"
        coord_name = name + "_dim_0"

        # Drop var/dim if already present
        if name in posterior.data_vars:
            posterior = posterior.drop_vars(name).drop_dims(coord_name)

        coords = ("chain", "draw", coord_name)
        posterior[name] = (coords, mean)
        posterior = posterior.assign_coords({coord_name: list(range(obs_n))})
        return posterior


class Categorical(MultivariateFamily):
    SUPPORTED_LINKS = ["softmax"]

    def predict(self, model, posterior, linear_predictor):
        # This is only 'softmax' for now.
        # Second axis is the one where the response coord is inserted
        # We need to append zeros for  the reference category
        shape = linear_predictor.shape
        linear_predictor = np.dstack(
            (np.zeros(shape=(shape[0], shape[1], 1, shape[3])), linear_predictor)
        )

        mean = self.link.linkinv(linear_predictor, axis=2)

        obs_n = mean.shape[-1]
        name = model.response.name + "_mean"
        coord_name = model.response.name + "_obs"

        # Drop var/dim if already present
        if name in posterior.data_vars:
            posterior = posterior.drop_vars(name).drop_dims(coord_name)

        response_coord_name = model.response.name + "_mean_coord"
        coords = ("chain", "draw", response_coord_name, coord_name)
        posterior[name] = (coords, mean)
        posterior = posterior.assign_coords({response_coord_name: model.response.levels})
        posterior = posterior.assign_coords({coord_name: list(range(obs_n))})
        return posterior



import pymc3 as pm


def get_distribution(dist):
    """Return a PyMC3 distribution."""
    if isinstance(dist, str):
        if hasattr(pm, dist):
            dist = getattr(pm, dist)
        else:
            raise ValueError(f"The Distribution '{dist}' was not found in PyMC3")
    return dist


def has_hyperprior(kwargs):
    """Determines if a Prior has an hyperprior"""
    return (
        "sigma" in kwargs
        and "observed" not in kwargs
        and isinstance(kwargs["sigma"], pm.model.TransformedRV)
    )



"""Classes to represent prior distributions and methods to set automatic priors"""
from .prior import Prior
from .scaler_mle import PriorScalerMLE
from .scaler_default import PriorScaler

__all__ = [
    "Prior",
    "PriorScaler",
    "PriorScalerMLE",
]



from .family import Family

# Names of parameters that can receive a prior distribution for the built-in families
FAMILY_PARAMS = {
    "beta": ("kappa",),
    "gamma": ("alpha",),
    "gaussian": ("sigma",),
    "negativebinomial": ("alpha",),
    "t": ("sigma", "nu"),
    "wald": ("lam",),
}


def _extract_family_prior(family, priors):
    """Extract priors for a given family

    If a key in the priors dictionary matches the name of a nuisance parameter of the response
    distribution for the given family, this function extracts and returns the prior for that
    nuisance parameter. The result of this function can be safely used to update the ``Prior`` of
    the response term.

    Parameters
    ----------
    family: str or ``bambi.families.Family``
        The family for which we want to extract priors.
    priors: dict
        A dictionary where keys represent parameter/term names and values represent
        prior distributions.
    """
    if isinstance(family, str) and family in FAMILY_PARAMS:
        names = FAMILY_PARAMS[family]
        priors = {name: priors.pop(name) for name in names if priors.get(name) is not None}
        if priors:
            return priors
    elif isinstance(family, Family):
        # Only work if there are auxiliary parameters in the family, and if any of these are
        # present in 'priors' dictionary.
        nuisance_params = list(family.likelihood.priors)
        if set(nuisance_params).intersection(set(priors)):
            return {k: priors.pop(k) for k in nuisance_params if k in priors}
    return None



"""Classes to construct model families."""
from .family import Family
from .likelihood import Likelihood
from .link import Link
from .utils import _extract_family_prior

__all__ = [
    "Family",
    "Likelihood",
    "Link",
]



# pylint: disable=no-name-in-module
import sys

from os.path import dirname, join

import numpy as np
import pandas as pd

from scipy.special import hyp2f1
from statsmodels.genmod.generalized_linear_model import GLM
from statsmodels.tools.sm_exceptions import PerfectSeparationError

from bambi.families.univariate import Gaussian, StudentT

from .prior import Prior


class PriorScalerMLE:
    """Scale prior distributions parameters.

    Used internally. Based on https://arxiv.org/abs/1702.01201
    """

    # Default is 'wide'. The wide prior sigma is sqrt(1/3) = .577 on the partial
    # corr scale, which is the sigma of a flat prior over [-1,1].
    names = {"narrow": 0.2, "medium": 0.4, "wide": 3 ** -0.5, "superwide": 0.8}

    def __init__(self, model, taylor):
        self.model = model

        # Equal to the design matrix for the common terms. Categorical are like "var[level]".
        # Q: What if the model does not have common effects? Not even intercepts. It doesn't work
        # right now. Should we flag it? Attempt to fix it?
        if model._design.common:
            self.dm = model._design.common.as_dataframe()
        else:
            self.dm = pd.DataFrame()

        self.has_intercept = model.intercept_term is not None

        self.priors = {}
        self.mle = None
        self.taylor = taylor
        # pylint: disable=unspecified-encoding
        with open(join(dirname(__file__), "config", "derivs.txt"), "r") as file:
            self.deriv = [next(file).strip("\n") for x in range(taylor + 1)]

    def get_intercept_stats(self, add_slopes=True):
        # Start with mean and variance of Y on the link scale
        mod = GLM(
            endog=self.model.response.data,
            exog=np.repeat(1, len(self.model.response.data)),
            family=self.model.family.smfamily(self.model.family.smlink),
            missing="drop" if self.model.dropna else "none",
        ).fit()
        mu = mod.params

        # Multiply SE by sqrt(N) to turn it into (approx.) sigma(Y) on link scale
        sigma = (mod.cov_params()[0] * len(mod.mu)) ** 0.5

        # Modify mu and sigma based on means and sigmas of slope priors.
        if self.model.common_terms and add_slopes:
            # prior["mu"] and prior["sigma"] have more than one value when the term is categoric.
            means = np.hstack([prior["mu"] for prior in self.priors.values()])
            sigmas = np.hstack([prior["sigma"] for prior in self.priors.values()])
            x_mean = np.hstack([self.model.terms[term].data.mean(axis=0) for term in self.priors])

            mu -= np.dot(means, x_mean)
            sigma = (sigma ** 2 + np.dot(sigmas ** 2, x_mean ** 2)) ** 0.5
        return mu, sigma

    def get_slope_stats(self, exog, name, values, sigma_corr, points=4, full_model=None):
        """
        Parameters
        ----------
        name: str
            Name of the term
        values: np.array
            Values of the term
        full_model: statsmodels.genmod.generalized_linear_model.GLM
            Statsmodels GLM to replace MLE model. For when ``'predictor'`` is not in the common
            part of the model.
        points : int
            Number of points to use for LL approximation.
        """

        # Make sure 'name' is in 'exog' columns

        if full_model is None:
            full_model = self.mle

        # Get log-likelihood values from beta=0 to beta=MLE
        beta_mle = full_model.params[name].item()
        beta_seq = np.linspace(0, beta_mle, points)

        log_likelihood = get_llh(self.model, exog, full_model, name, values, beta_seq)
        coef_a, coef_b = get_llh_coeffs(log_likelihood, beta_mle, beta_seq)
        p, q = shape_params(sigma_corr)

        # Evaluate the derivatives of beta = f(correlation).
        # dict 'point' gives points about which to Taylor expand.
        # We want to expand about the mean (generally 0), but some of the derivatives
        # do not exist at 0. Evaluating at a point very close to 0 (e.g., .001)
        # generally gives good results, but the higher order the expansion, the
        # further from 0 we need to evaluate the derivatives, or they blow up.
        point = dict(zip(range(1, 14), 2 ** np.linspace(-1, 5, 13) / 100))
        vals = dict(a=coef_a, b=coef_b, n=len(self.model.response.data), r=point[self.taylor])
        deriv = [eval(x, globals(), vals) for x in self.deriv]  # pylint: disable=eval-used

        terms = [
            compute_sigma(deriv, p, q, i, j)
            for i in range(1, self.taylor + 1)
            for j in range(1, self.taylor + 1)
        ]
        return np.array(terms).sum() ** 0.5

    def scale_response(self):
        # Add cases for other families
        priors = self.model.family.likelihood.priors
        if isinstance(self.model.family, (Gaussian, StudentT)):
            if priors["sigma"].auto_scale:
                sigma = np.std(self.model.response.data)
                priors["sigma"] = Prior("HalfStudentT", nu=4, sigma=sigma)

    def scale_common(self, term):
        """Scale common terms, excluding intercepts."""
        # Defaults are only defined for Normal priors
        if term.prior.name != "Normal":
            return
        mu = []
        sigma = []
        sigma_corr = term.prior.scale
        for name, values in zip(term.levels, term.data.T):
            mu += [0]
            sigma += [
                self.get_slope_stats(exog=self.dm, name=name, values=values, sigma_corr=sigma_corr)
            ]
        # Save and set prior
        self.priors.update({term.name: {"mu": mu, "sigma": sigma}})
        term.prior.update(mu=np.array(mu), sigma=np.array(sigma))

    def scale_intercept(self, term):
        # Default priors are only defined for Normal priors
        if term.prior.name != "Normal":
            return

        # Get prior mean and sigma for common intercept
        mu, sigma = self.get_intercept_stats()

        # Save and set prior
        term.prior.update(mu=mu, sigma=sigma)

    def scale_group_specific(self, term):
        # these default priors are only defined for HalfNormal priors
        if term.prior.args["sigma"].name != "HalfNormal":
            return

        sigma_corr = term.prior.scale

        # recreate the corresponding common effect data
        data_as_common = term.predictor

        # Handle intercepts
        if term.kind == "intercept":
            _, sigma = self.get_intercept_stats()
            sigma *= sigma_corr
        # Handle slopes
        else:
            # Check whether the expr is also included as common term in the model.
            expr = term.name.split("|")[0]
            term_levels_len = term.predictor.shape[1]

            # Handle case where there IS a corresponding common effect with same encoding
            if expr in self.priors and term_levels_len == len(self.priors[expr]["mu"]):
                sigma = self.priors[expr]["sigma"]
            # Handle case where there IS NOT a corresponding common effect
            else:
                if expr in self.priors and not term_levels_len == len(self.priors[expr]["mu"]):
                    # Common effect is present, but with different encoding
                    # Replace columns from the common term with those from the group specific term.
                    exog = self.dm.drop(self.model.terms[expr].levels, axis=1)
                else:
                    # Common effect is not present
                    exog = self.dm

                # Append columns from 'data_as_common'
                df_to_append = pd.DataFrame(data_as_common)
                df_to_append.columns = [f"_name_{i}" for i in df_to_append.columns]
                exog = exog.join(df_to_append)

                # If there's intercept and the term is cell means, drop intercept to avoid
                # linear dependence in design matrix columns.
                if term.is_cell_means and self.has_intercept:
                    exog = exog.drop("Intercept", axis=1)

                sigma = []
                for name, values in zip(df_to_append.columns, data_as_common.T):
                    full_model = GLM(
                        endog=self.model.response.data,
                        exog=exog,
                        family=self.model.family.smfamily(self.model.family.smlink),
                        missing="drop" if self.model.dropna else "none",
                    ).fit()
                    sigma += [
                        self.get_slope_stats(
                            exog=exog,
                            name=name,
                            values=values,
                            full_model=full_model,
                            sigma_corr=sigma_corr,
                        )
                    ]
                sigma = np.array(sigma)
        # Set the prior sigma.
        term.prior.args["sigma"].update(sigma=np.squeeze(np.atleast_1d(sigma)))

    def scale(self):
        # Classify all terms
        common = list(self.model.common_terms.values())
        group_specific = list(self.model.group_specific_terms.values())

        if self.has_intercept:
            intercept = [self.model.intercept_term]
        else:
            intercept = []

        # Arrange them in the order in which they should be initialized
        terms = common + intercept + group_specific
        term_types = (
            ["common"] * len(common)
            + ["intercept"] * len(intercept)
            + ["group_specific"] * len(group_specific)
        )

        # Scale response
        self.scale_response()

        # Initialize terms in order
        for term, term_type in zip(terms, term_types):
            # Only scale priors if term or model is set to be auto scaled.
            # By default, use "wide".
            if not term.prior.auto_scale:
                continue

            if term.prior.scale is None:
                term.prior.scale = "wide"

            # Convert scale names to floats
            if isinstance(term.prior.scale, str):
                term.prior.scale = self.names[term.prior.scale]

            if self.mle is None:
                self.fit_mle()

            # Scale term with the appropiate method
            getattr(self, f"scale_{term_type}")(term)

    def fit_mle(self):
        """Fits MLE of the common part of the model.

        This used to be called in the class instantiation, but there is no need to fit the GLM when
        there are no automatic priors. So this method is only called when needed.
        """
        missing = "drop" if self.model.dropna else "none"
        try:
            self.mle = GLM(
                endog=self.model.response.data,
                exog=self.dm,
                family=self.model.family.smfamily(self.model.family.smlink),
                missing=missing,
            ).fit()
        except PerfectSeparationError as error:
            msg = "Perfect separation detected, automatic priors are not available. "
            msg += "Please indicate priors manually."
            raise PerfectSeparationError(msg) from error
        except:
            msg = "Unexpected error when trying to compute automatic priors."
            msg += "Please indicate priors manually."
            print(msg, sys.exc_info()[0])
            raise


def get_llh(model, exog, full_model, name, values, beta_seq):
    """
    Parameters
    ---------
    model: bambi.Model
    exog: pandas.DataFrame
    name: str
        Name of the term for which we want to compute the llh
    values: np.array
        Values of the term for which we want to compute the llh
    beta_seq: np.array
        Sequence of values from to beta_mle.
    """
    # True if there are other predictors appart from `predictor_name`
    if name not in exog.columns:
        raise ValueError("get_llh failed. Term name not in exog.")

    multiple_predictors = exog.shape[1] > 1
    sm_family = model.family.smfamily(model.family.smlink)

    if multiple_predictors:
        # Use statsmodels to _optimize_ the LL. Model is fitted 'points' times.
        glm_model = GLM(endog=model.response.data, exog=exog, family=sm_family)
        null_models = [glm_model.fit_constrained(f"{name}={beta}") for beta in beta_seq[:-1]]
        null_models = np.append(null_models, full_model)
        log_likelihood = np.array([x.llf for x in null_models])
    else:
        # Use statsmodels to _evaluate_ the LL. Model is fitted 'points' times.
        log_likelihood = [
            sm_family.loglike(np.squeeze(model.response.data), beta * values)
            for beta in beta_seq[:-1]
        ]
        log_likelihood = np.append(log_likelihood, full_model.llf)
    return log_likelihood


def moment(p, q, k):
    """Return central moments of rescaled beta distribution"""
    return (2 * p / (p + q)) ** k * hyp2f1(p, -k, p + q, (p + q) / p)


def compute_sigma(deriv, p, q, i, j):
    """Compute and return the approximate sigma"""
    return (
        1
        / np.math.factorial(i)
        * 1
        / np.math.factorial(j)
        * deriv[i]
        * deriv[j]
        * (moment(p, q, i + j) - moment(p, q, i) * moment(p, q, j))
    )


def get_llh_coeffs(llh, beta_mle, beta_seq):
    # compute params of quartic approximation to log-likelihood
    # c: intercept, d: shift parameter
    # a: quartic coefficient, b: quadratic coefficient
    # beta_mle: beta obtained via MLE
    # beta_seq: sequence from 0 to beta_mle
    intercept, shift_parameter = llh[-1], -beta_mle
    X = np.array([(beta_seq + shift_parameter) ** 4, (beta_seq + shift_parameter) ** 2]).T
    a, b = np.squeeze(  # pylint: disable=invalid-name
        np.linalg.multi_dot([np.linalg.inv(np.dot(X.T, X)), X.T, (llh[:, None] - intercept)])
    )
    return a, b


def shape_params(sigma_corr, mean=0.5):
    # m, v: mean and variance of beta distribution of correlations
    # p, q: corresponding shape parameters of beta distribution
    mean = 0.5
    variance = sigma_corr ** 2 / 4
    p = mean * (mean * (1 - mean) / variance - 1)
    q = (1 - mean) * (mean * (1 - mean) / variance - 1)
    return p, q



from numpy.linalg import matrix_rank


def listify(obj):
    """Wrap all non-list or tuple objects in a list.

    Provides a simple way to accept flexible arguments.
    """
    if obj is None:
        return []
    else:
        return obj if isinstance(obj, (list, tuple, type(None))) else [obj]


def spacify(string):
    """Add 2 spaces to the beginning of each line in a multi-line string."""
    return "  " + "  ".join(string.splitlines(True))


def multilinify(sequence, sep=","):
    """Make a multi-line string out of a sequence of strings."""
    sep += "\n"
    return "\n" + sep.join(sequence)


def check_full_rank(matrix):
    """Checks if a matrix is full rank

    Parameters
    ----------
    matrix: numpy.array
        A 2-dimensional NumPy array that represents a design matrix

    Returns
    -------
    None
    """
    if matrix_rank(matrix) < matrix.shape[1]:
        raise ValueError(
            "Design matrix for common effects is not full-rank. "
            "Bambi does not support sparse settings when automatic priors are obtained via MLE."
        )



import numpy as np


class ResponseTerm:
    """Representation of a single response model term.

    Parameters
    ----------
    term: formulae.ResponseVector
        An object describing the response of the model,
        as returned by ``formulae.design_matrices().response``
    """

    def __init__(self, term):
        self.name = term.name
        self.data = term.design_vector
        self.constant = np.var(self.data) == 0  # NOTE: ATM we're not using this one
        self.categorical = term.type == "categoric"
        self.baseline = None  # Not None for non-binary categorical variables
        self.success = term.success if term.success is not None else 1  # not None for binary vars
        self.levels = None  # Not None for categorical variables
        self.binary = None  # Not None for categorical variables (either True or False)
        self.alias = None

        if self.categorical:
            self.binary = term.binary
            self.levels = term.levels
            if self.binary:
                self.success = term.success
            else:
                self.baseline = term.baseline

        if self.categorical:
            self.binary = term.binary
            self.levels = term.levels
            if self.binary:
                self.success = term.success
            else:
                self.baseline = term.baseline

        # We use pymc coords when the response is multi-categorical.
        # These help to give the appropriate shape to coefficients and make the resulting
        # InferenceData object much cleaner
        self.pymc_coords = {}
        if self.categorical and not self.binary:
            name = self.name + "_coord"
            self.pymc_coords[name] = term.levels[1:]

        # We use pymc coords when the response is multi-categorical.
        # These help to give the appropriate shape to coefficients and make the resulting
        # InferenceData object much cleaner
        self.pymc_coords = {}
        if self.categorical and not self.binary:
            name = self.name + "_coord"
            self.pymc_coords[name] = term.levels[1:]

    def set_alias(self, value):
        self.alias = value

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"shape: {self.data.squeeze().shape}",
        ]

        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"

        if self.categorical:
            args += [f"levels: {self.levels}"]
            if self.binary:
                args += [f"success: {self.success}"]
            else:
                args += [f"baseline: {self.baseline}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class Term:
    """Representation of a single (common) model term.

    Parameters
    ----------
    name : str
        Name of the term.
    term_dict: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().common.terms_info``
    data : ndarray
        The term values.
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    """

    group_specific = False

    def __init__(self, name, term_dict, data, prior=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.kind = term_dict["type"]
        self.levels = term_dict["full_names"]
        self.categorical = False
        self.term_dict = term_dict
        self.alias = None

        # If the term has one component, it's categorical if the component is categorical.
        # If the term has more than one component (i.e. it is an interaction), it's categorical if
        # at least one of the components is categorical.
        if self.kind == "interaction":
            if any(term["type"] == "categoric" for term in term_dict["terms"].values()):
                self.categorical = True
        else:
            self.categorical = self.kind == "categoric"

        # Flag constant terms
        if self.categorical and len(term_dict["levels"]) == 1 and (data == data[0]).all():
            raise ValueError(f"The term '{name}' has only 1 category!")

        if not self.categorical and self.kind != "intercept" and np.all(data == data[0]):
            raise ValueError(f"The term '{name}' is constant!")

        # Flag cell-means terms (i.e., full-rank coding), which receive special priors
        # To flag intercepts we use `self.kind`
        self.is_cell_means = self.categorical and (self.data.sum(1) == 1).all()

        # Obtain pymc coordinates, only for categorical components of a term.
        # A categorical component can have up to two coordinates if it is including with both
        # reduced and full rank encodings.
        self.pymc_coords = {}
        if self.categorical:
            name = self.name + "_coord"
            if self.kind == "interaction":
                self.pymc_coords[name] = term_dict["levels"]
            elif term_dict["encoding"] == "full":
                self.pymc_coords[name] = term_dict["levels"]
            else:
                self.pymc_coords[name] = term_dict["levels"][1:]

    def set_alias(self, value):
        self.alias = value

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"prior: {self.prior}",
            f"kind: {self.kind}",
            f"shape: {self.data.squeeze().shape}",
            f"categorical: {self.categorical}",
        ]

        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"

        if self.categorical:
            args += [f"levels: {self.levels}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()


class GroupSpecificTerm:
    # pylint: disable=too-many-instance-attributes
    """Representation of a single (group specific) model term.

    Parameters
    ----------
    name : str
        Name of the term.
    term: dict
        A dictionary describing the components of the term. These can be found in
        ``formulae.design_matrices().group.terms_info``
    data : (DataFrame, Series, ndarray)
        The term values.
    prior : Prior
        A specification of the prior(s) to use. An instance of class ``priors.Prior``.
    """

    group_specific = True

    def __init__(self, name, term, data, prior=None):
        self.name = name
        self.data = data
        self.prior = prior
        self.kind = term["type"]
        self.groups = term["groups"]
        self.levels = term["full_names"]
        self.grouper = term["Ji"]
        self.predictor = term["Xi"]
        self.group_index = self.invert_dummies(self.grouper)
        self.categorical = False
        self.term = term
        self.alias = None
        self.hyperprior_alias = {}

        # Determine if the expression is categorical
        if self.kind == "interaction":
            if any(t["type"] == "categoric" for t in term["terms"].values()):
                self.categorical = True
        else:
            self.categorical = self.kind == "categoric"

        # Determine if the term represents cell-means encoding.
        self.is_cell_means = self.categorical and (self.data.sum(1) == 1).all()

        # Used in pymc3 model coords to label coordinates appropiately
        self.pymc_coords = {}
        # Group is always a coordinate added to the model.
        expr, factor = self.name.split("|")
        self.pymc_coords[factor + "_coord_group_factor"] = self.groups

        if self.categorical:
            name = expr + "_coord_group_expr"
            if self.kind == "interaction":
                self.pymc_coords[name] = term["levels"]
            elif term["encoding"] == "full":
                self.pymc_coords[name] = term["levels"]
            else:
                self.pymc_coords[name] = term["levels"][1:]

    def invert_dummies(self, dummies):
        """
        For the sake of computational efficiency (i.e., to avoid lots of large matrix
        multiplications in the backend), invert the dummy-coding process and represent full-rank
        dummies as a vector of indices into the coefficients.
        """
        vec = np.zeros(len(dummies), dtype=int)
        for i in range(1, dummies.shape[1]):
            vec[dummies[:, i] == 1] = i
        return vec

    def set_alias(self, value):
        self.alias = value

    def set_hyperprior_alias(self, name, value):
        self.hyperprior_alias.update({name: value})

    def __str__(self):
        args = [
            f"name: {self.name}",
            f"prior: {self.prior}",
            f"groups: {self.groups}",
            f"type: {self.kind}",
            f"shape: {self.data.squeeze().shape}",
            f"categorical: {self.categorical}",
        ]

        if self.alias:
            args[0] = f"{args[0]} (alias: {self.alias})"

        if self.categorical:
            args += [f"levels: {self.levels}"]

        return f"{self.__class__.__name__}({', '.join(args)})"

    def __repr__(self):
        return self.__str__()
